/*
 * Copyright (C) 2009 Qi Hardware Inc.,
 * Author:  Xiangfu Liu <xiangfu@qi-hardware.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 3 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <p32mx/p32mx_core.h>
#include <p32mx/p32mx_clkc.h>
#include <p32mx/p32mx_uart.h>
#include <p32mx/p32mx_sys.h>
#include <p32mx/console.h>

#ifndef UART_CH
#define UART_CH	0
#endif

#define DEFAULT_BAUD	57600

static inline void serial_setbrg(const long baud) {
	U_BRG(UART_CH)  =  (clkc_get_pbclk()/16/baud - 1 ) & 0xffff;
}

static inline int serial_can_putc() {
	if (regx_is_set(U_STAh,  UART_CH , U_STA_UTXBF)) return 0;
	return 1;
}

static int __serial_putc (char c, FILE *fp)
{
	unsigned long flags;

	if (c == '\n') __serial_putc ('\r', fp);
	while (!serial_can_putc())
		;
	flags = __irq_disable();
	U_TXREG(UART_CH) = c;
	__irq_restore(flags);
	return 0;	// 0 or EOF
}

#if (UART_CH == 0)
void serial0_init(void)
#else
void serial1_init(void)
#endif
{
    FILE *fp;
#if defined(__32MX2XX__) && (UART_CH == 0)
    RPB4R = PPS1_OUT_U1TX;
    U1RXR = PPS3_IN_RPA4;
#elif defined(__32MX2XX__) && (UART_CH == 1)
    RPB9R = PPS4_OUT_U2TX;
    U2RXR = PPS2_IN_RPB8;
#endif
    U_MODE(UART_CH) = (1<<U_MODE_ON);
    U_STAh(UART_CH) = (1<<U_STA_UTXEN) | (2<<U_STA_UTXISEL)
		    | (1<< U_STA_URXEN) | (0<<U_STA_URXISEL);
    serial_setbrg(DEFAULT_BAUD);

    fp = fdevopen(__serial_putc, NULL);
    jzcon_append(fp);
}

#if 0
FILE *fdevopen_serial (FILE *fp)
{
	serial_init();
	if (fp) {
		fdev_setup_stream(fp, __serial_putc, NULL, _FDEV_SETUP_WRITE);
	} else {
		fp = fdevopen(__serial_putc, NULL);
	}
	return fp;
}
#endif
